/*****************************************************************************/

/* 

Code for computing object pose using POSIT 

Inputs: 1) a file containing number of points and coordinates of object points;
        2) a file containing integer coordinates of image points;
Important: Image point at position i in image file should correspond to object point at 
position i in object file

Outputs: 1) Rotation of scene with respect to camera
         2) Translation vector from projection center of camera
	    to FIRST point of scene (object) file.
            
Reference: D. DeMenthon and L.S. Davis, "Model-Based Object Pose in 25 Lines of Code", 
International Journal of Computer Vision, 15, pp. 123-141, June 1995. 

*/

/*****************************************************************************/

#include <stdio.h>
#include <math.h>
#include <stdlib.h>

/*****************************************************************************/

const int maxCount = 30;/* exit iteration with a message after this many loops */
const int nbObjectCoords = 3; /* x, y, z */
const int nbImageCoords = 2; /* x, y */

/*****************************************************************************/

/* Structures */

typedef struct TObject{
	int nbPts;
	double **objectPts;/* Array of coordinates of points */
	double **objectVects;/* Array of coordinates of vectors from reference pt to all pts */
	double **objectCopy;/* Copy of objectVects, used because SVD code destroys input data */
	double **objectMatrix;/* Pseudoinverse of objectVects */
}TObject;

typedef struct TImage{
	int nbPts;
	int imageCenter[2];
	int **imagePts;
	double **imageVects;/* scaled orthographic projections */
	double **oldImageVects;/* projections at previous iteration step */
	double *epsilon;/* Corrections to scaled orthographic projections at each iteration */
}TImage;

typedef struct TCamera{
	int focalLength;
	double rotation[3][3];/* Rotation of SCENE in camera reference, NOT other way around */
	double translation[3];/* Translation of SCENE in camera reference */
}TCamera;

/*****************************************************************************/

/* Prototypes: */

double **InitDoubleArray(int nbRows, int nbCols);
int **InitIntArray(int nbRows, int nbCols);
void FreeDoubleArray(double **anArray, int nbRows);
void FreeIntArray(int **anArray, int nbRows);
void PrintMatrix(double **mat, int nbRows, int nbCols);
void PrintVector(double vect[], int nbElems);
double **ReadObject(char *name, int *nbPts);
int **ReadImage(char *name, int imageCenter[2], int nbPts);
void GetObjectAndImage(char *objectName, char *imageName, TObject *object, TImage *image);
void POS(TObject object, TImage image, TCamera *camera);
void POSIT(TObject object, TImage image, TCamera *camera);
long GetImageDifference(TImage image);
int  PseudoInverse(double **A,int N, double **B);
void PrintRotation(double rotation[][3]);
void nrerror(char error_text[]);
int main();

/*****************************************************************************/


int debug1 = 1;/* 0 for no printout */

/*****************************************************************************/

double **ReadObject(char *name, int *nbPts)
/* 
Read object feature points from a file 
*/
{
FILE *fp;
int i, j;
float coord;
double **objectPts;
		
	fp = fopen(name, "r");
	if (fp != NULL){
                if(debug1) printf("Object data read from file:\n");
		fscanf(fp, "%d", nbPts);
		if(debug1) printf("%6d\n", *nbPts);
		
		/* Initialize objectPts array */
		objectPts = InitDoubleArray(*nbPts, nbObjectCoords);
		
		for (i=0; i< *nbPts;i++){
 			for (j=0; j<nbObjectCoords; j++){
				fscanf(fp, "%f", &coord);
				objectPts[i][j] = coord;
				if(debug1) printf("%.2f ",coord);
			}
			if(debug1) printf("\n");
		}
		if(debug1) printf("\n");
	}
	else nrerror("could not open file of object points");
	return objectPts;
}

/*****************************************************************************/

int **ReadImage(char *name, int imageCenter[2], int nbPts)
/*
read coords of image points, which are INTEGER
there should be as many image points as there were object points
and the order should be the same order as that of the corresponding 
object points of the file read by ReadObject.
*/
{
FILE *fp;
int i, j;
int coord;
int **imagePts;
	
	fp = fopen(name, "r");
	if (fp != NULL){
		imagePts = InitIntArray(nbPts, nbImageCoords);
                if(debug1) printf("Image data read from file:\n");
		for (i=0; i<nbPts;i++){
			for (j=0; j<2; j++){
				fscanf(fp, "%d", &coord);
				imagePts[i][j] = coord - imageCenter[j];
				if(debug1) printf("%d ", coord);
			}
			if(debug1) printf("\n");
		}
		if(debug1) printf("\n");
	}
	else printf("could not open file ");
	return imagePts;
}

/*****************************************************************************/

void GetObjectAndImage(char *objectName, char *imageName, TObject *object, TImage *image)
/* 
Initialize data structures for object and image data and read data from files 
*/
{
int i, j, nbPoints;
/* Object data */

	object->objectPts = ReadObject(objectName, &nbPoints);
	object->nbPts = nbPoints;

	object->objectVects = InitDoubleArray(object->nbPts, nbObjectCoords);
	object->objectCopy = InitDoubleArray(object->nbPts, nbObjectCoords);
		
	for (i=0;i<object->nbPts;i++){
		for(j=0;j<nbObjectCoords;j++){
			object->objectVects[i][j] = object->objectCopy[i][j] =
					object->objectPts[i][j] - object->objectPts[0][j];
		}
	}

	object->objectMatrix = InitDoubleArray(nbObjectCoords, object->nbPts);

/* Image data */
	
	image->nbPts = object->nbPts;
	image->imagePts = ReadImage(imageName, image->imageCenter, image->nbPts);

	image->imageVects = InitDoubleArray(image->nbPts, nbImageCoords);
	image->oldImageVects = InitDoubleArray(image->nbPts, nbImageCoords);
	image->epsilon = (double *)malloc(image->nbPts * sizeof(double));
}

	
/*****************************************************************************/

void POS(TObject object, TImage image, TCamera *camera)
/*
POS function 
(Pose from Orthography and Scaling, a scaled orthographic proj. approximation).
Returns one translation and one rotation.
*/
{
double    I0[3], J0[3], row1[3], row2[3], row3[3];
double    I0I0, J0J0;
int  i, j;
double scale, scale1, scale2;


	/*Computing I0 and J0, the vectors I and J in TRs listed above */
	for (i=0;i<nbObjectCoords;i++){
    	I0[i]=0;
    	J0[i]=0;
    	for (j=0;j<object.nbPts;j++){
			I0[i]+=object.objectMatrix[i][j]*image.imageVects[j][0];
			J0[i]+=object.objectMatrix[i][j]*image.imageVects[j][1];
    	}
	}

	I0I0=I0[0]*I0[0] + I0[1]*I0[1] + I0[2]*I0[2];
	J0J0=J0[0]*J0[0] + J0[1]*J0[1] + J0[2]*J0[2];

	scale1 = sqrt(I0I0);
	scale2 = sqrt(J0J0);
	scale = (scale1 + scale2) / 2.0;
	
	/*Computing TRANSLATION */
	camera->translation[0] = image.imagePts[0][0]/scale;
	camera->translation[1] = image.imagePts[0][1]/scale;
	camera->translation[2] = camera->focalLength/scale;

	/* Computing ROTATION */
	for (i=0;i<3;i++){
 	   row1[i]=I0[i]/scale1;
 	   row2[i]=J0[i]/scale2;
	}
	row3[0]=row1[1]*row2[2]-row1[2]*row2[1];/* Cross-product to obtain third row */
	row3[1]=row1[2]*row2[0]-row1[0]*row2[2];
	row3[2]=row1[0]*row2[1]-row1[1]*row2[0];
	for (i=0;i<3;i++){ 
 	   camera->rotation[0][i]=row1[i];
 	   camera->rotation[1][i]=row2[i];
 	   camera->rotation[2][i]=row3[i];
	}
}

/*****************************************************************************/

long GetImageDifference(TImage image)
/*
Get sum of differences between coordinates in lists
of old image points and new image points
*/
{
int i, j;
long sumOfDiffs = 0;
	
	for (i=0;i<image.nbPts;i++){
		for (j=0;j<2;j++){
			sumOfDiffs += abs(floor(0.5+image.imageVects[i][j])-floor(0.5+image.oldImageVects[i][j]));
		}
	}
	return sumOfDiffs;
}

/*****************************************************************************/

void POSIT(TObject object, TImage image, TCamera *camera)
/* 
Iterate over results obtained by the POS function;
see paper "Model-Based Object Pose in 25 Lines of Code", IJCV 15, pp. 123-141, 1995.
*/
{
int i, j, iCount;
int converged;
long imageDiff;

/* Starting point for iteration loop */
	for(iCount=0;iCount<maxCount;iCount++){
		if(iCount==0){
			for (i=0;i<image.nbPts;i++){
				for(j=0;j<nbImageCoords;j++){
   				image.imageVects[i][j]=(double)(image.imagePts[i][j]-image.imagePts[0][j]);
				}
   	 	}
   	 }
   	 else{/* iCount>0 */
			/* Compute new imageVects */
			for (i=0;i<image.nbPts;i++){
				image.epsilon[i] = 0.0;
				for (j=0;j<3;j++){
 	  				image.epsilon[i] += object.objectVects[i][j] * camera->rotation[2][j]; /*dot product M0Mi.k*/
 	  			}
 	  			image.epsilon[i] /= camera->translation[2]; /* divide by Z0 */
 	  		}
 			/* Corrected image vectors */	
			for (i=0;i<image.nbPts;i++){
				for(j=0;j<nbImageCoords;j++){
	    			image.imageVects[i][j]=(double)image.imagePts[i][j]*(1+image.epsilon[i])-image.imagePts[0][j];
				}
 			}
			imageDiff = GetImageDifference(image);/*using pts gives same result */
			printf("imageDiff %ld\n",imageDiff);
		}
			
		/* Remember old imageVects */
		for (i=0;i<image.nbPts;i++){
    		image.oldImageVects[i][0]=image.imageVects[i][0];
   	 	image.oldImageVects[i][1]=image.imageVects[i][1];
		}
		
		/* Compute rotation and translation */
		POS(object,image, camera);
	
 		converged = (iCount>0 && imageDiff==0);
 		if(converged) break;

		if(iCount==maxCount && !converged){
			printf("POSIT did not converge \n");
		}
	}/*end for*/;
}

/*****************************************************************************/

void PrintRotation(double rotation[][3])
{
int i, j;

	for (i=0; i<3;i++){
		for (j=0; j<3; j++){
			printf("%.3f ", rotation[i][j]);
		}
		printf("\n");
	}
	printf("\n");
}

/*****************************************************************************/




int main()
/*
Read file of image points, file of object points,
compute object matrix by pseudoinverse operation
then get the pose
*/
{
int fLength = 760;/* Focal Length in PIXELS */
int imageCenterX = 0;/* In this test, origin of image coordinates is at image center */
int imageCenterY = 0;/* More often, imageCenterX=256, imageCenterY=256 */
int isMaxRank;
TObject testObject;
TImage testImage;
TCamera testCamera;

	testImage.imageCenter[0] = imageCenterX;
	testImage.imageCenter[1] = imageCenterY;

	/* Init structures and read object and image data:*/	
	GetObjectAndImage("TestCube", "CubeImage", &testObject, &testImage);
		
	testCamera.focalLength = fLength;

	/* Find object matrix as pseudoInverse of matrix of object vector coordinates */	
	isMaxRank = PseudoInverse(testObject.objectCopy, testObject.nbPts, testObject.objectMatrix);
	/* Check if matrix rank for object matrix is less than 3: */
	if(!isMaxRank){
		nrerror("object is too flat; another method is required ");/* exit */
	}
	else{
		/* Find object pose by POSIT, Pose from Orthography, Scaling, and ITerations */
		POSIT(testObject, testImage, &testCamera);

		/* Print results */
		printf("\n");
		printf("Rotation matrix of scene in camera reference frame\n");
		PrintRotation(testCamera.rotation);
		printf("\n");
		printf("Translation vector of scene in camera reference frame\n 
                        from camera projecction center to FIRST point of scene (object) file\n");
		PrintVector(testCamera.translation, nbObjectCoords);
		printf("\n");
	}
	/* Do a bit of memory cleanup: */
	FreeDoubleArray(testObject.objectPts, testObject.nbPts);
	FreeDoubleArray(testObject.objectVects, testObject.nbPts);
	FreeDoubleArray(testObject.objectCopy, testObject.nbPts);
	FreeDoubleArray(testObject.objectMatrix, nbObjectCoords);
	FreeIntArray(testImage.imagePts, testImage.nbPts);
	FreeDoubleArray(testImage.imageVects, testImage.nbPts);
	FreeDoubleArray(testImage.oldImageVects, testImage.nbPts);
	
        return 0;	
}

